/* outpaulb.c	output routines for the Netwide Assembler to produce
 *		a debugging trace
 *
 * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
 * Julian Hall. All rights reserved. The software is
 * redistributable under the licence given in the file "Licence"
 * distributed in the NASM archive.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "nasm.h"
#include "nasmlib.h"
#include "outform.h"

struct Section {
    struct Section *next;
    long number;
    char *name;
} *dbgsect;

FILE *dbgf;
efunc dbgef;

static int origin_defined;
static unsigned long	cur_offset = 0;
static int need_offset_output = 1;

struct ofmt of_paulb;


static const char *paulb_stdmac[] = {
   "%define __SECT__ [section .text]",
   "%imacro org 1+.nolist",
   "[org %1]",
   "%endmacro",
   "%macro __NASM_CDecl__ 1",
   "%endmacro",
   NULL
};


static void paulb_init(FILE *fp, efunc errfunc, ldfunc ldef, evalfunc eval)
{
    (void) eval;

    dbgf = fp;
    dbgef = errfunc;
    dbgsect = NULL;
    (void) ldef;
    fprintf(fp,"NASM Output format debug dump\n");
    of_paulb.current_dfmt->init(&of_paulb,0,fp,errfunc);
    
}

static void paulb_cleanup(int debuginfo)
{
    (void) debuginfo;
    of_paulb.current_dfmt->cleanup();
    while (dbgsect) {
	struct Section *tmp = dbgsect;
	dbgsect = dbgsect->next;
	nasm_free (tmp->name);
	nasm_free (tmp);
    }
    fclose(dbgf);
}

static long paulb_section_names (char *name, int pass, int *bits)
{
    int seg;

    /*
     * We must have an initial default: let's make it 16.
     */
    if (!name)
	*bits = 16;

    if (!name)
	fprintf(dbgf, "section_name on init: returning %d\n",
		seg = seg_alloc());
    else {
	int n = strcspn(name, " \t");
	char *sname = nasm_strndup(name, n);
	struct Section *s;

	seg = NO_SEG;
	for (s = dbgsect; s; s = s->next)
	    if (!strcmp(s->name, sname))
		seg = s->number;
	
	if (seg == NO_SEG) {
	    s = nasm_malloc(sizeof(*s));
	    s->name = sname;
	    s->number = seg = seg_alloc();
	    s->next = dbgsect;
	    dbgsect = s;
	    fprintf(dbgf, "section_name %s (pass %d): returning %d\n",
		    name, pass, seg);
	}
    }
    return seg;
}

static void paulb_deflabel (char *name, long segment, long offset,
			  int is_global, char *special) 
{
#if 0
    fprintf(dbgf,"deflabel %s := %08lx:%08lx %s (%d)%s%s\n",
	    name, segment, offset,
	    is_global == 2 ? "common" : is_global ? "global" : "local",
	    is_global,
	    special ? ": " : "", special);
#endif
}


static void paulb_out (long segto, void *data, unsigned long type,
		     long segment, long wrt) 
{
    long realbytes = type & OUT_SIZMASK;
    long ldata;
    int id;

    type &= OUT_TYPMASK;

#if	0
fprintf(dbgf, "# 0x%8.8x\n", cur_offset);
cur_offset += realbytes;
 
   fprintf(dbgf,"out to %lx, len = %ld: ",segto,realbytes);

    switch(type) {
      case OUT_RESERVE:
	fprintf(dbgf,"reserved.\n"); break;
      case OUT_RAWDATA:
	fprintf(dbgf,"raw data = ");
	while (realbytes--) {
	    id = *(unsigned char *)data;
	    data = (char *)data + 1;
	    fprintf(dbgf,"%02x ",id);
	}
	fprintf(dbgf,"\n"); break;
      case OUT_ADDRESS:
	ldata = 0; /* placate gcc */
	if (realbytes == 1)
	    ldata = *((char *)data);
	else if (realbytes == 2)
	    ldata = *((short *)data);
	else if (realbytes == 4)
	    ldata = *((long *)data);
	fprintf(dbgf,"addr %08lx (seg %08lx, wrt %08lx)\n",ldata,
		segment,wrt);break;
      case OUT_REL2ADR:
	fprintf(dbgf,"rel2adr %04x (seg %08lx)\n",(int)*(short *)data,segment);
	break;
      case OUT_REL4ADR:
	fprintf(dbgf,"rel4adr %08lx (seg %08lx)\n",*(long *)data,segment);
	break;
      default:
	fprintf(dbgf,"unknown\n");
	break;
    }
#else

	if (need_offset_output) {
		fprintf(dbgf, "=%8.8x\n", cur_offset);
		need_offset_output = 0;
	}
	cur_offset += realbytes;

	switch(type) {
	case OUT_RESERVE:
		need_offset_output = 1;
		break;
	case OUT_RAWDATA:
	case OUT_ADDRESS:
		while(realbytes--) {
			id = *(unsigned char *)data;
			data = (char *)data + 1;
			fprintf(dbgf, "%2.2x ", id);
		}
		fprintf(dbgf, "\n");
		break;
	case OUT_REL2ADR:
		fprintf(dbgf, "OUT_REL2ADR\n");
		break;
	case OUT_REL4ADR:
		fprintf(dbgf, "OUT_REL4ADR\n");
		break;
	default:
		break;
	}
#endif

}

static long paulb_segbase(long segment) 
{
    return segment;
}

static int paulb_directive (char *directive, char *args, int pass) 
{
   /* Handle ORG directive */
   if (!nasm_stricmp(directive, "org"))
   {  struct tokenval tokval;
      unsigned long value;
      expr *e;

      stdscan_reset();
      stdscan_bufptr = args;
      tokval.t_type = TOKEN_INVALID;
      e = evaluate(stdscan, NULL, &tokval, NULL, 1, dbgef, NULL);
      if (e)
      {  if (!is_really_simple(e))
            dbgef(ERR_NONFATAL, "org value must be a critical"
               " expression");
         else
         {  value = (unsigned long)reloc_value(e);
            if (value < 0)
               dbgef(ERR_NONFATAL, "attempt to set a negative program origin");
            /* Check for ORG redefinition. */
            else if (origin_defined && (value != cur_offset))
               dbgef(ERR_NONFATAL, "program origin redefined");
            else
            {  cur_offset = value;
               origin_defined = 1;
		need_offset_output = 1;
            }
         }
      }
      else
         dbgef(ERR_NONFATAL, "No or invalid offset specified"
            " in ORG directive.");
      return 1;
   }

    return 1;
}

static void paulb_filename (char *inname, char *outname, efunc error) 
{
    standard_extension (inname, outname, ".paulb", error);
}

static int paulb_set_info(enum geninfo type, char **val)
{
    (void) type;
    (void) val;
    return 0;
}
char *paulb_types[] = { 
	"unknown", "label", "byte","word","dword","float","qword","tbyte" 
};
void paulb_dbg_init(struct ofmt * of, void * id, FILE * fp, efunc error)
{
    (void) of;
    (void) id;
    (void) fp;
    (void) error;
    fprintf(fp,"   With debug info\n");
}
static void paulb_dbg_cleanup(void)
{
}

static void paulb_dbg_linnum (const char *lnfname, long lineno, long segto)
{
#if 0
    fprintf(dbgf,"dbglinenum %s(%ld) := %08lx\n",
	lnfname,lineno,segto);
#endif
}
static void paulb_dbg_deflabel (char *name, long segment,
			  long offset, int is_global, char *special) 
{
#if 0
    fprintf(dbgf,"dbglabel %s := %08lx:%08lx %s (%d)%s%s\n",
	    name,
            segment, offset,
	    is_global == 2 ? "common" : is_global ? "global" : "local",
	    is_global,
	    special ? ": " : "", special);

#endif
}
static void paulb_dbg_define(const char *type, const char *params)
{
    fprintf(dbgf,"dbgdirective [%s] value [%s]\n",type, params);
}
static void paulb_dbg_output (int output_type, void *param)
{
    (void) output_type;
    (void) param;
}
static void paulb_dbg_typevalue(long type)
{
	fprintf(dbgf,"new type: %s(%lX)\n",
	    paulb_types[TYM_TYPE(type) >> 3], TYM_ELEMENTS(type) );
}
static struct dfmt debug_debug_form = {
    "Trace of all info passed to debug stage",
    "debug",
    paulb_dbg_init,
    paulb_dbg_linnum,
    paulb_dbg_deflabel,
    paulb_dbg_define,
    paulb_dbg_typevalue,
    paulb_dbg_output,
    paulb_dbg_cleanup,
};

static struct dfmt *debug_debug_arr[3] = {
	&debug_debug_form,
	&null_debug_form,
	NULL
};
struct ofmt of_paulb = {
    "Trace of all info passed to output stage",
    "paulb",
    NULL,
    debug_debug_arr,
    &null_debug_form,
    paulb_stdmac,
    paulb_init,
    paulb_set_info,
    paulb_out,
    paulb_deflabel,
    paulb_section_names,
    paulb_segbase,
    paulb_directive,
    paulb_filename,
    paulb_cleanup
};
